# coding=utf8
import io
import sys
import hashlib
import time
from pathlib import Path
import os
import shutil
import datetime
import glob


class fileHandler:
    """
    Класс работы с файлами для загрузки
    """
    BUF_SIZE = 65536
    SLEEP_TIME_SEC = 3
    processed_folder = "SUCCESS"
    error_folder = "FAILED"
    FILE = ""

    def __init__(self, file, mask = ""):
        """
        Конструктор, вызываемый в конструкции with () as ...
        """
        root_folder = Path(file).parents[0]

        if mask != '':
            try:
                mask_path = os.path.join(root_folder, mask)
                files = glob.glob(mask_path)
                files.sort(key = os.path.getmtime)
                file = files[0]
            except Exception as e:
                print ('Couldn''t find files according to mask')
                raise  Exception (e)

        self.FILE = file
        file_name = Path(file).name

        self.processed_folder = os.path.join(root_folder, self.processed_folder, file_name)
        # print(self.processed_folder)
        self.mkdir(Path(self.processed_folder).parents[0])

        self.error_folder = os.path.join(root_folder, self.error_folder, file_name)
        # print(self.error_folder)
        self.mkdir(Path(self.error_folder).parents[0])


        if self.is_file_complete(file):
            pass
            # self.move_to_processed()
        else:
            raise Exception ("File is incomplete- can't load it just yet")
            # TODO exiting- file isn't ready just yet

        # if self.is_already_loaded(file):
        #     pass
        #     #TODO are we doing something in case of duplicate load?

    def mkdir(self, dir):
        ret = ""
        try:
            if not os.path.exists(dir):
                os.mkdir(dir)
        except Exception as e:
            ret = e
        return ret

    def get_inboxed_file(self):
        pass

    def move_to_inboxed(self):
        pass

    def move_to_processed(self):
        self.move_to(self.FILE, self.processed_folder)

    def move_to_error(self):
        self.move_to(self.FILE, self.error_folder)

    def move_to(self, source, destination):
        shutil.move(source, destination)

    def is_file_complete(self, file):
        hash1 = self.get_hash(file)
        time.sleep(self.SLEEP_TIME_SEC)
        if self.get_hash(file) == hash1:
            is_complete = True
        else:
            is_complete = False
        return is_complete

    def is_already_loaded(self, file):
        hash_new = self.get_hash(file)
        hash_old = self.get_hash(self.processed_folder)
        if hash_new == hash_old:
            is_already_loaded = True
        else:
            is_already_loaded = False
        return is_already_loaded

    def get_hash(self, file):
        sha1 = hashlib.sha512()
        with open(file, 'rb') as f:
            while True:
                data = f.read(self.BUF_SIZE)
                if not data:
                    break
                sha1.update(data)
        return sha1.hexdigest()

# def main():
#     with fileHandler('/Users/dio/Downloads/vertica_loader/num_test/ord_export.csv') as fh:
#         pass
#
# main()
# hashlib.sha512(str(dt.encode('utf8'))).hexdigest()