# coding=utf8
import vertica_python as dbif
import pandas as pd
import numpy as np
import logging
import errno
import datetime
import sys
import csv
from os import environ
import re
import os.path
from os import walk, remove
import io
from ukd_vertica_loader.ddlWriter import ddlWriter
from ukd_vertica_loader.dataInput import dataInput
from base64 import b64decode
from ukd_vertica_loader.fileHandler import fileHandler
from ukd_vertica_loader.blocker import blocker

# logging.basicConfig(level=logging.DEBUG,
#                     format='%(asctime)s : %(levelname)s : %(message)s',
#                     datefmt='%a, %d %b %Y %H:%M:%S',
#                     filename=os.path.join('/Users/artemiy/Downloads/sibur/ETL/Data/001-liquidity/', datetime.datetime.now().strftime('%Y-%m-%d') + '.log'),
#                     filemode='w')

class verticaLoader:
    """
    Класс загрузчика в HP Vertica
    """
    log_path = ""
    temp_dir = ""
    output_format = ""
    encoding = "utf-8"
    error_flag_name = ""
    connector = ""
    quotechar = '~'
    dateformat = '%Y-%m-%d %H:%M:%S.%f'
    lineterminator = '\036'
    is_numbered_col = False
    # lineterminator = "E''"
    rejected_table_name = 'STG_VERTICA_LOADER.REJ_DATA'
    __default_source_path = "/Users/dio/Downloads/vertica_loader"
    Blocker = ''

    __CONN_FILE_LOCATION = os.path.join(
        os.path.dirname(__file__),
        'connections.json'
    )

    __CONF_FILE_LOCATION = os.path.join(
        os.path.dirname(__file__),
        'verticaLoaderConf.json'
    )
    __LOAD_REJECT_FILEPATH = 'rejectedOnLoad.csv'
    __DEFAULT_DB_TAG = 'cloud'

    V_COPY_DELIMITER = '\035'
    # __V_COPY_ROWSEP    = '\\n'
    V_COPY_NULL = '\001'
    v_type_mapper = {'FLOAT': 'str', 'INT': 'float64', 'INTEGER': 'float64', 'DOUBLE': 'str', 'MONEY': 'str',
                     'DATE': 'str', 'TIME': 'str', 'TIMESTAMP': 'str', 'BOOLEAN': 'bool', 'TIMESTAMPZ': 'str'}

    v_metadata_columns = ['_FILENAME_', '_LOAD_DATETIME_']

    __tabname_cre = re.compile(r'([\w]+\.)')

    from ukd_vertica_loader.verticaLoaderSQLs import SQL

    __GET_CONSTR_SQL = SQL['GET_CONSTR']
    __GET_COLS_SQL = SQL['GET_COLS']
    __GET_SEGMENTATION_SQL = SQL['GET_SEGMENTATION']
    __GET_SORT_ORDER_SQL = SQL['GET_SORT_ORDER']
    __CREATE_TEMP_TABLE_SQL = SQL['CREATE_TEMP_TABLE']
    __MERGE_TABLES_SQL = SQL['MERGE_TABLES']

    def __get_filename_from_path(self, file):
        return file.split(os.path.sep)[-1]

    def __get_path_from_filename(self, file):
        return os.path.join(*file.split(os.path.sep)[:-1])

    def __cleanup_folder(self, path):
        files = self.get_file_list(path, "")
        for file in files:
            full_path = file
            # remove(full_path)

    def __map_pandas_type (self, vertica_type):
        vertica_type = vertica_type.upper()
        if 'VARCHAR' in vertica_type:
            ret = 'str'
        elif 'NUMERIC' in vertica_type:
            ret = 'str'
        else:
            ret = self.v_type_mapper.get(vertica_type)
        return ret

    def __load_source_from_file(self, full_path):
        conf = self.__parse_source_config(full_path)
        payload = 'dummy'
        self.__write_meta_load(payload)

    def __write_meta_load(self, payload):
        pass

    def __get_file_list(self, path, extension):
        path = os.path.normpath(path)
        file_list = []
        for (dirpath, dirnames, filenames) in walk(path):
            for file in filenames:
                if file.lower().endswith(extension):
                    full_file = os.path.join(dirpath, file)
                    file_list.append(full_file)
        return file_list

    def __parse_source_config (self, full_path):
        conf = json.loads(open(full_path, encoding = 'utf-8'))
        logging.debug (conf)
        return conf

    def check_folder(self, input_path = __default_source_path, data_path = '', fmd_path = ''):
        config_format = '.json'

        conf_list = self.__get_file_list(input_path, config_format)
        logging.debug (conf_list)
        if len(conf_list) == 0:
            logging.info('No files to load. Check verticaLoaderConf.json and ensure path is correct')
            print (('No files to load. Check verticaLoaderConf.json and ensure path is correct'))
        for config_file in conf_list:
            try:
                source_conf = self.__getSourceConfig(config_file)
                if self.Blocker.is_blocked(config_file):
                    logging.error('Error! File is being loaded already')
                    continue
                else:
                    self.Blocker.block_file(config_file, '')
                for k, v in source_conf.items():
                    self.__loadStageFile(v, data_path, fmd_path)
                self.Blocker.free_file(config_file)
            except Exception as e:
                logging.error("Error during metadata file {config_file} loading {error}".format(config_file=config_file, error=str(e)))
                self.Blocker.free_file(config_file)

        return

    def getSettings(self, connections= None, connFilepath=__CONN_FILE_LOCATION, confFilepath = __CONF_FILE_LOCATION) -> dict:
        """
        Вспомогательный метод, считывающий настройки из json'a
        NOTE: сейчас путь к json задан фиксированно
        """

        import json
        from base64 import b64decode

        # logger = self.__configure_logger()
        logging.info("getSettings(): Reading settings...")

        if connections == None:
            airflow_mode = 0
        else:
            airflow_mode = 1

        if airflow_mode == 0:
            # Gotta read config from connnections.json
            settings = self.getConnectionSettings(connFilepath)
            settings['common'] = self.getSettingsCommon()
        elif airflow_mode == 1:
            # Getting connection params from airflow module
            settings = self.getConnectionAirflow(connections)
            settings['common'] = connections.get("common", {})
        else:
            logging.error("Unknown connection type- I only support connections.json or airflow")
            # Probably should throw some error- no idea where to get the desired connection parameters



        logging.info("getSettings(): Done")
        return settings

    def getConnectionAirflow(self, connections) -> dict:
        ret_con = {}
        sn = connections["connections"].get('source_name', 'default')
        ret_con[sn] = connections.get("connections", None)


        for sn, sprop in ret_con.items():
            utype = sprop.get('userType', 'explicit')
            ptype = sprop.get('passwordType', 'explicit')

            if utype == "env":
                ret_con[sn]['user'] = environ.get(sprop['userEnvVar'])
            else:
                ret_con[sn]['user'] = sprop['user']

            # здесь ещё есть ветка, позволяющая явно прописывать пароль
            # в base64, но этот метод плох. Решили использовать переменные
            # среды для передачи этой информации
            if ptype == "env":
                ret_con[sn]['password'] = environ.get(sprop['passwordEnvVar'])
            elif ptype == "explicit":
                if sprop.get("passwordEncryption", 1) == 1:
                    ret_con[sn]['password'] = b64decode(
                        sprop["password"].encode('ascii')
                    ).decode('utf-8')
                else:
                    ret_con[sn]['password'] = sprop["password"]

        return ret_con

    def getConnectionSettings(self, connFilepath=__CONN_FILE_LOCATION) -> dict:
        """
        Вспомогательный метод, считывающий настройки из json'a
        NOTE: сейчас путь к json задан фиксированно
        """

        import json

        # logger = self.__configure_logger()
        logging.info("getConnectionSettings(): Reading settings...")

        # читаем JSON с настройками подключений к источникам
        with io.open(connFilepath, encoding = 'utf-8') as fp:
            json_text = fp.read()
            settings = json.loads(json_text)

        # Обрабатываем свойства
        for sn, sprop in settings.items():
            utype = sprop.get('userType', 'explicit')
            ptype = sprop.get('passwordType', 'explicit')

            if utype == "env":
                settings[sn]['user'] = environ.get(sprop['userEnvVar'])
            else:
                settings[sn]['user'] = sprop['user']

            # здесь ещё есть ветка, позволяющая явно прописывать пароль
            # в base64, но этот метод плох. Решили использовать переменные
            # среды для передачи этой информации
            if ptype == "env":
                settings[sn]['password'] = environ.get(sprop['passwordEnvVar'])
            elif ptype == "explicit":
                if sprop.get("passwordEncryption", 1) == 1:
                    settings[sn]['password'] = b64decode(
                        sprop["password"].encode('ascii')
                    ).decode('utf-8')
                else:
                    settings[sn]['password'] = sprop["password"]

        logging.info("getConnectionSettings(): Done")
        return settings

    def getSettingsCommon(self, confFilepath = __CONF_FILE_LOCATION) -> dict:
        import json
        from base64 import b64decode

        # logger = self.__configure_logger()
        logging.info("getSettings(): Reading settings...")

        # читаем JSON с настройками подключений к источникам
        with io.open(confFilepath, encoding = 'utf-8') as fp:
            json_text = fp.read()
            settings = json.loads(json_text)

        logging.info("getSettings(): Done")
        return settings

    def __init__(self, connections=None, settings=None):
        """
        Конструктор, вызываемый в конструкции with verticaLoader() as ...
        """

        self.dbtag = self.__DEFAULT_DB_TAG

        # Загрузка конфигурации
        if settings is None:
            if connections is None:
                self.smap = self.getSettings()
            else:
                self.smap = self.getSettings(connections=connections)
        elif isinstance(settings, dict):
            self.smap = settings
        elif isinstance(settings, str):
            self.smap = self.getSettings(settings)
        logging.debug (self.smap)

        self.vconn = None  # connection pointer
        # blocker_path = self.smap['common'].get('blocker_path', 'db.json').lower()
        blocker_path = os.path.join(self.smap['common']['temp_dir'], 'db.json')
        self.Blocker = blocker(blocker_path)
        logging.info(blocker_path)


    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.vconn is not None:
            del self.vconn
            self.vconn = None

    ##
    # CONNECTION MANAGEMENT
    ##

    def __getConnectionParameters(self, source_system) -> dict:
        """
        Возвращает словарь с параметрами соединения
        """

        logger = self.configure_logger()
        logger.info('__getConnectionParameters: call')
        return {
            'host': self.smap[source_system]['host'],
            'port': 5433,
            'user': self.smap[source_system]['user'],
            'password': self.smap[source_system]['password'],
            'database': self.smap[source_system]['db'],
            'read_timeout': 600,
            # default throw error on invalid UTF-8 results
            'unicode_error': 'strict',
            # SSL is disabled by default
            'ssl': False
        }
        # OLD Pyodbc variant
        # return "DRIVER={Vertica};SERVER=" + self.smap[dnORup]['host'] + \
        #                         ";DATABASE=" + self.smap[dnORup]['db'] + \
        #                         ";user=" + self.smap[dnORup]['user'] + \
        #                         ";password=" + self.smap[dnORup]['password']

    ##
    ## TODO: connection once acquired doesn't checked to be alive. Consider revision
    def __getConnection(self, source_system) -> 'connection':
        """
        Возвращает соединение с СУБД. Если оно удалось.
        """

        logger = self.configure_logger()
        logger.info('__getConnection(): call')
        if self.vconn is None:
            try:
                pars = self.__getConnectionParameters(source_system)
                self.vconn = dbif.connect(**pars)
            except Exception as e:
                logging.error("Error during connection creation:", sys.exc_info()[1])
                # rethrow?
        return self.vconn
    #
    # LOW LEVEL MANAGEMENT
    #

    def __getCopyCommandSql(self, columns: list, targetTable: str) -> str:

        logger = self.configure_logger()

        res = """COPY {table}
                 ({header})
                 FROM LOCAL STDIN
                 DELIMITER '{delim}'
                 NULL '{nulv}'
                 TRAILING NULLCOLS
                 ENCLOSED BY '{quotechar}'
                 RECORD TERMINATOR '{lineterminator}'
                 ABORT ON ERROR
                 """
        resf = res.format(
            table=targetTable,
            header=",".join(columns),
            quotechar=self.quotechar,
            delim=self.V_COPY_DELIMITER,
            lineterminator=self.lineterminator,
            # sep=self.__V_COPY_ROWSEP,
            nulv=self.V_COPY_NULL,
            rejfp=self.__LOAD_REJECT_FILEPATH)
        logger.debug('__getCopyCommandSql: SQL\n {}'.format(resf))
        return resf

    def __getVsqlCopyCommand(self, columns: list, targetTable: str) -> str:
        vtmpl = """vsql -U {user} -w {password} -h {host} -d {db} -c """
        сtmpl = "\"{copysql}\";"
        logging.debug("__getVsqlCopyCommand() call")
        res = \
            vtmpl.format(**(self.smap)) + \
            сtmpl.format(
                copysql=self.__getCopyCommandSql(columns, targetTable))
        return res


    def __loadDfToVertica(self, source_system, pdDf: pd.DataFrame, targetTable: str, use_vsql=False):
        """
        Основная функция загрузки, позволяет выбрать интерфейс вручную
        """
        cols = pdDf.columns

        if use_vsql:
            # use vsql piped stream
            vsqlCommand = self.__getVsqlCopyCommand(cols, targetTable)
            self.__loadDfToVerticaWithVsql(source_system, pdDf, vsqlCommand)
        else:
            if self.rejected_table_name != None:
                RejectedString = """REJECTED DATA AS TABLE {rej_table} """.format(rej_table=self.rejected_table_name)
            else:
                RejectedString = ""
            # use vertica_python copy from serialized csv
            copyCommand = """COPY {table} ({header}) FROM STDIN 
                        RECORD TERMINATOR '{lineterminator}' DELIMITER '{delim}' NULL as '{nullvalue}' 
                        {rej_string} """.format(
                table=targetTable,
                header=",".join(cols),
                delim=self.V_COPY_DELIMITER,
                quotechar=self.quotechar,
                lineterminator=self.lineterminator,
                nullvalue=self.V_COPY_NULL,
                rej_string=RejectedString
            )
            logging.info(copyCommand)
            self.__loadDfToVerticaWithCopy(source_system, pdDf, copyCommand, targetTable)

    def loadAndMerge(self, data: pd.DataFrame, targetTable: str, source_system: str) -> None:
        """
        Загружает данные в таблицу путём слияния данных по ключам.
        Дупликация по ключам не происходит, но новые данные полностью перетирают старые.
        """
        logging.debug('loadAndMerge(): Start load process')
        # get columns information

        temptable = self.__createAndLoadTempTable(source_system, data, targetTable)

        self.__merge(temptable, targetTable, source_system)

        self.runSql("DROP TABLE {}".format(temptable), source_system)


        # perform load to vertica
        logging.debug('loadAndMerge(): done')

    def loadAndMergeDf(self, data: pd.DataFrame, targetTable: str, source_system: str, structure: pd.DataFrame) -> None:
        pass
        #TODO
        schema_name = targetTable.split('.')[0]
        table = targetTable.split('.')[1]
        groups = self.convertMetaDF(structure, table, schema_name)
        if self.__compareDDL(source_system, table, groups):
            self.loadAndMerge(data, targetTable, source_system)
        else:
            pass
            #spotted structure inconsistency

    def convertMetaDF(self, input_df, target_table, schema_name) -> pd.DataFrame:
        name_mapper = {'Имя в БД': 'Column', 'Тип данных': 'DataType', 'Ключ (для поиска дублей)': 'PK', 'Обязательно': 'NotNull'}
        input_df.rename(mapper=name_mapper, inplace=True)
        input_df['Table'] = target_table
        input_df['Schema'] = schema_name
        return input_df

    def __createAndLoadTempTable(self, source_system: str, pdDf: pd.DataFrame, targetTable: str) -> str:
        """
        Создаёт временную таблицу (действительна до закрытии сессии) и загружает в неё данные.
        Возвращает наименование таблицы.
        """
        schema, table = targetTable.split(".")

        # TODO
        # first - obtain segmentation clause
        segmClause = self.__getSuperProjSegmentationClause(source_system, table, schema)
        if segmClause is None:
            segmClause = 'UNSEGMENTED'
        else:
            segmClause = 'SEGMENTED BY ' + segmClause
        # second - get order clause
        try:
            orderClause = ",".join(self.__getSuperProjOrderClause(source_system, table, schema))
        except Exception as e:
            logging.error ('error during order clause generation: %s' % str(e))

        # construct temp table
        temptable = 'temp_' + table
        try:
            createSql = self.__CREATE_TEMP_TABLE_SQL.format(
                temptable=temptable,
                schema=schema,
                table=table,
                segmClause=segmClause,
                orderClause=orderClause
            )
            self.runSql(createSql, source_system)
        except Exception as e:
            logging.error ("Error temp table %s" % (str(e)))

        # load data there
        self.__loadDfToVertica(source_system, pdDf, temptable, use_vsql=False)

        return temptable
        # self.__mergeByKeys(tablefrom, tableto, PKlist)

    def __getSuperProjSegmentationClause(self, source_system, table: str, schema=None) -> str:
        """
        Возвращает выражение, использованное для сегментации любой из
        супер-проекций
        """

        if schema is None:
            schema, table = table.split(".")

        segmClause = self.runSqlFetchList(source_system,
            self.__GET_SEGMENTATION_SQL.format(
                schema=schema,
                table=table
            )
        )
        # In vertica 8.1 segmentation clause might store table name
        # it should be removed.
        if segmClause[0][0] is not None:
            res = self.__tabname_cre.sub('', segmClause[0][0])
        else:
            res = None



        return None

    def __getSuperProjOrderClause(self, source_system, table: str, schema=None) -> str:
        """
        Возвращает выражение, использованное для сортировки полей любой из
        проекций
        """

        if schema is None:
            schema, table = table.split(".")

        ordClause = self.runSqlFetchList(source_system,
            self.__GET_SORT_ORDER_SQL.format(
                schema=schema,
                table=table
            )
        )
        return ordClause[0]

    def __merge(self, incTable: str, targetTable: str, source_system: str):
        """
        Осуществляет слияние данных из инкремента в targetTable по ключам.
        Существенно использует то, что при построении временной таблицы использовались
        наименования и порядок столбцов целевой таблицы
        """

        # """MERGE
        # INTO {targetTable} t
        # USING {incTable} i
        # ON ({joinPKcond})
        # WHEN MATCHED THEN UPDATE SET
        #     {updateClause}
        # WHEN NOT MATCHED THEN INSERT ({colsCommaSep})
        #     VALUES ({icolsCommaSep})
        # """
        # Step 1: get metadata from database
        logging.debug("__merge(): call")
        # schema, table = targetTable.split(".")

        PKlist = self.__getPKColumns(source_system, targetTable)
        COlist = self.__getTableColumns(source_system, targetTable)

        logging.debug("PK: {}\n COL: {}".format(PKlist, COlist))
        pars = {
            'targetTable': targetTable,
            'incTable': incTable,
            'joinPKcond': " AND ".join(["t.{col}=i.{col}".format(col=c) for c in PKlist]),
            'updateClause': ",".join(["{col}=i.{col}".format(col=c) for c in COlist]),
            'colsCommaSep': ",".join(COlist),
            'icolsCommaSep': "i." + ",i.".join(COlist)
        }

        mergeSql = self.__MERGE_TABLES_SQL.format(**pars)
        self.runSql(mergeSql, source_system)

    ##
    # METADATA ANALYSIS
    ##

    def __getPKColumns(self, source_system: str, table: str, schema=None) -> list:
        """
        Возвращает список с атрибутами, которые являются первичными ключами
        """
        if schema is None:
            schema, table = table.split(".")

        PKlol = self.runSqlFetchList(source_system,
            self.__GET_CONSTR_SQL.format(
                schema=schema,
                table=table
            )
        )

        return [r[0] for r in PKlol]

    def runSqlFetchList(self, source_system: str, sqlCommand: str) -> list:
        logging.debug("runSqlFetchList(): \n{}".format(sqlCommand))
        conn = self.__getConnection(source_system)
        curs = conn.cursor()
        res = []

        try:
            curs.execute(sqlCommand)
            for row in curs.iterate():
                res.append(row)
        except Exception as e:
            logging.error(repr(e))
        curs.close()
        return res

    def runSqlFetchListParam(self, source_system: str, sqlCommand: str, params: list) -> list:
        logging.debug("runSqlFetchList(): \n{}".format(sqlCommand))
        conn = self.__getConnection(source_system)
        curs = conn.cursor()
        res = []

        try:
            curs.execute(sqlCommand, params)
            for row in curs.iterate():
                res.append(row)
        except Exception as e:
            logging.error(repr(e))
        curs.close()
        return res

    def __getTableColumns(self, source_system, table: str, schema=None) -> list:
        """
        Возвращает список атрибутов таблицы в порядке расположения
        """
        if schema is None:
            schema, table = table.split(".")

        ColsLol = self.runSqlFetchList(source_system,
            self.__GET_COLS_SQL.format(
                schema=schema,
                table=table
            )
        )

        return [r[0] for r in ColsLol]

    def truncateAndLoadDf(self, source_system: str, data: pd.DataFrame, targetTable: str, control: dict, is_truncate = True):
        #TODO
        logger = self.configure_logger()
        logger.info('loadStage : START : Getting ready-made dataframe from outside source')
        self.truncateAndLoad(source_system, data, targetTable, is_truncate)

    def truncateAndLoad(self, source_system: str, data: pd.DataFrame, targetTable: str, is_truncate = True):
        """
        Загружает данные в таблицу, перед этим удаляя старую версию
        """
        logging.info('TruncateAndLoad(): Start load process')

        if is_truncate == True:
            self.runSql("TRUNCATE TABLE {}".format(targetTable), source_system)
        self.__loadDfToVertica(source_system, data, targetTable)
        # perform load to vertica
        logging.info('truncateAndLoad(): done')

    def __loadDfToVerticaWithVsql(self, pdDf: pd.DataFrame, vsqlCommand: str):
        from subprocess import Popen, PIPE, STDOUT

        logger = self.configure_logger()
        logger.info("__loadDfToVerticaWithVsql(): call")
        try:
            with Popen(
                    vsqlCommand,
                    stdin=PIPE,
                    stdout=PIPE,
                    stderr=STDOUT, encoding='utf-8') as p:
                ctr = 0
                for row in pdDf.itertuples():
                    # SLOPPY, should be revised
                    p.stdin.write((self.V_COPY_DELIMITER).join(list(map(str, row[1:]))) + '\n')
                    ctr += 1
                    if (ctr % 100000 == 0):
                        p.stdin.flush()
                p.stdin.flush()
                logging.info('Load done')
                stdout = p.communicate('\\.\n')
                p.kill()
        except BrokenPipeError as e:
            logging.error("Error:", repr(e))


    def __loadDfToVerticaWithCopy(self, source_system, pdDf: pd.DataFrame, copyCommand: str, targetTable: str):
        from os import remove, path

        logger = self.configure_logger()
        logger.info("__loadDfToVerticaWithCopy(): call")
        tempfilename = os.path.join(self.temp_dir, targetTable + 'temp_copy.csv')
        logging.debug (pdDf.dtypes)
        try:
            conn = self.__getConnection(source_system)
            curs = conn.cursor()
            pdDf.to_csv(
                tempfilename,
                sep=self.V_COPY_DELIMITER,
                quoting=csv.QUOTE_NONE,
                # quotechar=self.quotechar,
                doublequote=False,
                line_terminator=self.lineterminator,
                index=False,
                header=False,
                date_format=self.dateformat,
                na_rep=self.V_COPY_NULL,
                escapechar='\\',
                float_format='%.0f',
                encoding='utf-8'
            )

            with open(tempfilename, 'r', encoding=self.encoding) as f:
                curs.copy(copyCommand, f, buffer_size=65536)

            try:
                num_rejected_rows = self.__readSql(source_system, """select get_num_rejected_rows();""")
                num_accepted_rows = self.__readSql(source_system, """select get_num_accepted_rows();""")
                logging.info("Number of accepted rows: %s" % num_accepted_rows)
                print("Number of accepted rows: %s" % num_accepted_rows)
                logging.info("Number of rejected rows: %s" % num_rejected_rows)
                print("Number of rejected rows: %s" % num_rejected_rows)
            except:
                logging.info('Logging and info output failed, but everything seems alright otherwise')
            conn.commit()

        except Exception as e:
            logging.error ("Connection error during copy: %s" % repr(e))
            raise ConnectionError("Connection error during copy: %s" % repr(e))
        finally:
            if path.isfile(tempfilename):
                # remove(tempfilename)
                pass
            curs.close()
        logger.info("__loadDfToVerticaWithCopy(): exit")

    def runSql(self, sqlCommand: str, source_system: str):

        logger = self.configure_logger()
        logger.debug("runSql(): \n{}".format(sqlCommand))

        conn = self.__getConnection(source_system)
        curs = conn.cursor()
        try:
            curs.execute(sqlCommand)
        except Exception as e:
            logging.error(repr(e))
            logging.error ('run sql error: %s' % str(e))
        try:
            curs.close()
        except:
            logging.error("Couldn't close connection")

    def readSqlDf(self, sqlCommand: str, source_system: str, params: list):
        conn = self.__getConnection(source_system)
        return pd.read_sql_query(sqlCommand,con = conn, params=params)

    def runSqlParam(self, sqlCommand: str, source_system: str, params: list):

        logger = self.configure_logger()
        logger.debug("runSql(): \n{}".format(sqlCommand))

        conn = self.__getConnection(source_system)
        curs = conn.cursor()
        try:
            curs.execute(sqlCommand, params)
        except Exception as e:
            logging.error(repr(e))
            logging.error ('run sql error: %s' % str(e))
        try:
            curs.close()
        except:
            logging.error("Couldn't close connection")

    def __readSql(self, source_system: str, sqlCommand: str) -> str:
        ret = ''

        logger = self.configure_logger()
        logger.debug("runSql(): \n{}".format(sqlCommand))

        conn = self.__getConnection(source_system)
        curs = conn.cursor()
        try:
            curs.execute(sqlCommand)
            ret = curs.fetchall()
        except Exception as e:
            logging.error(repr(e))
        try:
            curs.close()
        except:
            logging.error("Couldn't close connection")
        return ret

    def __getSourceConfig(self, file) -> dict:
        """
        метод читает json, описывающий загружаемые данные
        """
        import json

        # logger = self.configure_logger()
        logging.info("getSourceConfig(): Reading metadata...")

        with open(file, encoding='utf8') as fp:
            json_text = fp.read()
            source = json.loads(json_text)

        logging.info("getSourceConf(): Done")
        return source

    def loadStage(self):
        source_system = 'cloud'
        if type(self.smap['common'].get('paths', '')) == type('some_string'):

            self.check_folder(self.smap['common']['path'])
        else:
            for key, path_set in self.smap['common']['paths'].items():
                self.check_folder(path_set.get('path', ''), path_set.get('data_path', ''), path_set.get('fmd_path', ''))

    def __loadStageFile(self, config, data_path = '', fmd_path = ''):
        # Загрузка в таблицы STG
        file = config['file_name']
        file_list = file.split(',')
        source_system = config['source_system']
        fmd = config['fmd_file_name']
        method = config['method']
        fmd_sheets = config['fmd_sheets']
        sheets = config.get('sheets', 0)
        file_type = config.get('file_type', 'xlsx')
        csv_delimiter = config.get('csv_delimiter', ';')

        filename_attr = config.get('filename_attr', 'FILE')
        load_datetime_attr = config.get('load_datetime_attr', 'TIME')

        if data_path == '':
            data_path = self.smap['common']['data_path']

        if fmd_path == '':
            fmd_path = self.smap['common']['fmd_path']

        mask = config.get('file_mask_format', '')


        try:
            if sheets == '*':
                sheets = 0
        except:
            logging.debug('Nothing to do here, just here to not fail if type ')

        self.rejected_table_name = self.smap[source_system].get('rejected_table_name',
                                                                    self.smap['common'].get('rejected_table_name'))

        self.V_COPY_NULL = config.get('null_value', self.V_COPY_NULL)

        self.quotechar = config.get('quotechar', self.quotechar)


        skiprows = config.get('skiprows', -1)
        columns = config.get('read_columns', None)
        encoding = config.get('encoding', 'utf8')
        self.encoding = encoding

        logger = self.configure_logger()
        ddl_dir = os.path.join(self.smap['common']['log_dir'], 'ddl')
        self.temp_dir = os.path.join(self.smap['common']['temp_dir'], 'temp')
        self.__create_directory_if_not_exists(self.temp_dir)
        self.__create_directory_if_not_exists(ddl_dir)
        dwr = ddlWriter(os.path.join(ddl_dir, 'ddl_{file_name}.sql'.format(file_name=file)))

        logger.info('loadStage : START : Read metadata from FMD file')



        df = pd.read_excel(os.path.join(fmd_path, fmd), fmd_sheets,
                           index_col=None)
        df['Column'] = df['Column'].str.upper()
        df['Table'] = df['Table'].str.upper()
        groups = df.groupby('Table')
        pd.options.mode.chained_assignment = None
        tables = df['Table'].unique().tolist()


        logger.info('loadStage : FINISH : Read metadata from FMD file')

        for tab in tables:
            try:
                use_columns = None

                gr = groups.get_group(tab)
                cur_file = gr.file_source.values[0]
                if cur_file in file_list:
                    print (cur_file)
                    print (tab)
                    logger.info('loadStage : START : Compare DDL with data {t}'.format(t=tab))
                    fn, stmt = self.__prepareDDL(tab, groups)

                    logger.info('loadStage : FINISH : Compare DDL with data {t}'.format(t=tab))

                    logger.info('loadStage : START : Read data file sheet={t}'.format(t=tab))

                    if str(gr.file_attribute_name.values[0]) == 'nan':
                        # no header in excel- column order defined by attribute number in fmd
                        # no header in excel- column order defined by attribute number in fmd
                        gr.sort_values(by="file_attribute_number", ascending=True, inplace=True)
                        attr_list = gr.Column.values
                        attr_list = attr_list.tolist()
                        self.is_numbered_col = True
                    else:
                        # normal mapping (attributes to column names)
                        self.is_numbered_col = False
                        gr.sort_values(by="file_attribute_name", ascending=True, inplace=True)
                        use_columns = gr.file_attribute_name.values
                        use_columns = [x for x in use_columns if isinstance(x, str)]

                    column_mapper = {}
                    type_mapper = {}
                    timestamp_name = ''
                    filename_name = ''
                    dtype_dict = {}

                    self.v_metadata_columns = []
                    for index, row in gr.iterrows():
                        try:
                            cur_sheet = row.get('Sheet', '')
                            if cur_sheet == '' or cur_sheet!=cur_sheet:
                                cur_sheet = 0
                        except:
                            cur_sheet = 0
                        if str(row['is_metadata']) == '1.0':
                            if filename_attr.lower() in row['Column'].lower():
                                filename_name = row['Column']
                                self.v_metadata_columns.append(row['Column'].upper())
                                if self.is_numbered_col:
                                    attr_list.remove(row['Column'])
                            if load_datetime_attr.lower() in row['Column'].lower() or 'date' in row['Column'].lower():
                                timestamp_name = row['Column']
                                self.v_metadata_columns.append(row['Column'].upper())
                                if self.is_numbered_col:
                                    attr_list.remove(row['Column'])
                        elif sheets == cur_sheet or file_type == 'csv' or cur_sheet == 0:
                            column_mapper[row['file_attribute_name']] = row['Column'].upper()
                            type_mapper[row['Column']] = row['DataType']
                            if self.is_numbered_col == False:
                                dtype_dict[row['file_attribute_name']] = self.__map_pandas_type(row['DataType'])
                            else:
                                dtype_dict[int(row['file_attribute_number'])-1] = self.__map_pandas_type(row['DataType'])
                    if len(dtype_dict) == 0:
                        continue

                    file_loader = dataInput()
                    try:
                        if mask != '':
                            file_handler = fileHandler(os.path.join(data_path, file), mask)

                        else:
                            file_handler = fileHandler(os.path.join(data_path, file))
                        file = file_handler.FILE
                    except Exception as e:
                        logging.error("File is incomplete still- skipping it for now: {error}".format(error = e))
                        print ('File is incomplete still- skipping it for now: {error}'.format(error = e))
                        continue


                    if self.is_numbered_col == True:
                        if file_type == 'xlsx' or file_type == 'xls':
                            data = file_loader.read_excel_numbered(os.path.join(data_path, file), sheets,
                                                 attr_list, dtype_dict, skiprows,
                                                 columns)
                        elif file_type == 'csv':
                            data = file_loader.read_csv_numbered(
                                os.path.join(data_path, file),
                                attr_list, dtype_dict, skiprows, csv_delimiter, columns=attr_list, encoding=encoding)
                        else:
                            logging.error('Error! Unsupported file type')
                    else:
                        if file_type == 'xlsx' or file_type == 'xls':
                            data = file_loader.read_excel_named(os.path.join(data_path, file), sheets,
                                                 dtype_dict, skiprows, columns)
                        elif file_type == 'csv':
                            data = file_loader.read_csv_named(
                                os.path.join(data_path, file),
                                dtype_dict, skiprows, csv_delimiter, list(dtype_dict.keys()),columns=use_columns, encoding=encoding)
                        else:
                            logging.error('Error! Unsupported file type')

                    if self.is_numbered_col == False:
                        # checking if name mapper has duplicate column names
                        gr['file_attribute_name'] = gr['file_attribute_name'].fillna('')
                        if any(gr['file_attribute_name'][gr['file_attribute_name'] != ''].duplicated()):
                            logging.error('file attribute names contain duplicate values. Use numbering instead')
                            continue
                        data.rename(columns=column_mapper, inplace=True)

                        # checking if mapping went correctly
                        attr_list = gr.Column.values
                        attr_list = attr_list.tolist()
                        for col_name in attr_list:
                            if col_name.upper() not in data.columns and col_name.upper() not in self.v_metadata_columns:
                                logging.error('Seems that {s} hasn''t been mapped correctly'.format(s=col_name))
                                raise Exception('Seems that {s} hasn''t been mapped correctly'.format(s=col_name))

                    logging.debug(data.dtypes)
                    for col, dtype in type_mapper.items():
                        if 'TIMESTAMP' in dtype.upper():
                            try:
                                data[col] = pd.to_datetime(data[col], errors='raise')
                            except ValueError as e:
                                logging.error("Cant convert column %s to datetime due to %s" % (col, e))
                                raise Exception("Cant convert column %s to datetime due to %s" % (col, e))
                        if 'DATE' == dtype.upper():
                            try:
                                data[col] = pd.to_datetime(data[col], errors='raise')
                            except ValueError as e:
                                logging.error("Cant convert column %s to datetime due to %s" % (col, e))
                                raise Exception("Cant convert column %s to datetime due to %s" % (col, e))
                        if 'INT' in dtype.upper():
                            try:
                                data[col] = data[col].astype(float)
                            except ValueError as e:
                                logging.error("Cant convert column %s to int due to %s" % (col, e))
                                raise Exception("Cant convert column %s to int due to %s" % (col, e))
                        if dtype.upper() in ('FLOAT', 'DOUBLE', 'NUMERIC', 'DECIMAL', 'MONEY'):
                            try:
                                data[col] = data[col].fillna(self.V_COPY_NULL).astype(str)
                            except ValueError as e:
                                logging.error("Can't escape backlash in varchar due to %s" % e)
                                raise Exception("Can't escape backlash in varchar due to %s" % e)
                        if 'VARCHAR' in dtype.upper():
                            try:
                                logging.debug('Column %s fillna''d as str' % col)
                                data[col].fillna(self.V_COPY_NULL, inplace=True)
                                # data[col] = data[col].astype(str).replace('nan', self.V_COPY_NULL)
                            except Exception as e:
                                logging.error("Can't escape backlash in varchar due to %s" % e)
                                raise Exception("Can't escape backlash in varchar due to %s" % e)

                        if filename_name != '':
                            data[filename_name] = file
                        if timestamp_name != '':
                            data[timestamp_name] = pd.datetime.now().replace(microsecond=0)

                    logger.info('loadStage : START : loading comparison{t}'.format(t=tab))
                    logger.info("DDL for current table: %s" % stmt)

                    dwr.write(stmt)

                    if self.__compareDDL(source_system, tab, groups) == False:
                        logging.error('DDL comparison is unsuccessful. Not continuing with loading file')
                        continue

                    logger.info('loadStage : FINISH : Read data file sheet={t}'.format(t=tab))

                    logger.info('loadStage : START : Load data file into  {t}'.format(t=tab))
                    if method == 'snapshot':
                        logging.debug(data.dtypes)
                        self.truncateAndLoad(source_system, data, fn)
                        logger.info('loadStage : FINISH : Load data file into  {t} rows={n}'.format(t=tab, n=data.shape[0]))
                    elif method == 'merge':
                        self.loadAndMerge(data, fn, source_system)
                    elif method == 'insert':
                        self.truncateAndLoad(source_system, data, fn, False)

                    if self.smap['common'].get('move_loaded_files', 'true').lower() == 'true':
                        file_handler.move_to_processed()
            except Exception as e:
                logging.error('Error during table loading %s' % str(e))
                if self.smap['common'].get('move_loaded_files', 'true').lower() == 'true':
                    file_handler.move_to_error()
        pd.options.mode.chained_assignment = 'warn'

    def __prepareDDL(self, tab: pd.DataFrame, groups: pd.DataFrame) -> str:
        gr = groups.get_group(tab)

        # Gather full name
        schema = gr.Schema.values[0]
        fn = schema + '.' + tab

        # Gather COLUMN definitions
        gr['NotNull'] = np.where((gr.NotNull.isnull()), '', ' NOT NULL')
        atr = gr.Column + ' ' + gr.DataType + gr.NotNull.astype('str')
        attributes = ',\n'.join(atr.tolist())

        # Gather PK
        pk = ', '.join(gr[gr['PK'].notnull()].Column.tolist())
        pk = ', PRIMARY KEY (' + pk + ') \n' if len(pk) != 0 else ''

        # Get ORDER BY clause
        order_by = ', '.join(gr[gr['OrderBy'].notnull()].sort_values(by=['OrderBy']).Column.tolist())
        order_by = 'ORDER BY ' + order_by + ' \n' if len(order_by) != 0 else ''

        # Get SEGMENTED BY clause
        segmented_by = ', '.join(
            gr[gr['SegmentedBy'].notnull()].sort_values(by=['SegmentedBy']).Column.tolist())
        segmented_by = 'SEGMENTED BY HASH(' + segmented_by + ') ALL NODES' if len(
            segmented_by) != 0 else 'UNSEGMENTED ALL NODES'

        # Prepare SQL Statement
        stmt = "CREATE TABLE IF NOT EXISTS {fn} ({a} {p}) {o} {sg}; TRUNCATE TABLE {fn};".format(
            fn=fn,
            a=attributes,
            p=pk,
            o=order_by,
            sg=segmented_by)
        return fn, stmt

    def __compareDDL(self, source_system, tab: pd.DataFrame, groups: pd.DataFrame) -> str:
        allGood = True
        gr = groups.get_group(tab)
        # Gather full name
        schema = gr.Schema.values[0]
        fn = schema + '.' + tab

        # Gather COLUMN definitions
        gr['NotNull'] = np.where((gr.NotNull.isnull()), '', ' NOT NULL')
        atr = gr.Column + ' ' + gr.DataType + gr.NotNull.astype('str')
        attributes = ',\n'.join(atr.tolist())

        get_db_command = '''SELECT upper(column_name), upper(data_type) FROM v_catalog.columns
        where table_schema like %s and table_name like %s '''
        params = [schema, tab]

        db_state_df = self.readSqlDf(sqlCommand=get_db_command,source_system=source_system, params=params)
        db_state_df.columns = ['Column', 'DataType']
        gr.Column = gr.Column.astype('str')
        gr.DataType = gr.DataType.astype('str')

        excel_col_types = gr[['Column', 'DataType']]

        try:
            types_intersect = excel_col_types.merge(db_state_df, how='left', on='Column')
            error_cols = types_intersect[types_intersect.isna().any(1)]
            if len(error_cols.index) > 0:
                logging.error('Columns {col_name} dont exist in the database'.format(col_name=error_cols))
                allGood = False
        except Exception as e:
            logging.error('Error during ddl comparison {err}'.format(err=e))
            allGood = False

        return allGood


    def configure_logger(self):
        """
        Declare and validate existence of log directory; create and configure logger object
        :return:  instance of configured logger object
        """
        log_dir = os.path.join(self.smap['common']['log_dir'], 'log')
        self.__create_directory_if_not_exists(log_dir)#\l;'
        log_level = self.smap['common'].get('log_level', 'error').lower()
        if log_level == 'info':
            log_level = logging.INFO
        elif log_level == 'debug':
            log_level = logging.DEBUG
        elif log_level == 'error':
            log_level = logging.ERROR

        self.__configure_logging(log_dir, log_level)
        logger = logging.getLogger('importer_logger')
        return logger

    def __create_directory_if_not_exists(self, path):
        """
        Creates 'path' if it does not exist
        If creation fails, an exception will be thrown
        :param logger:  the logger
        :param path:    the path to ensure it exists
        """
        try:
            os.makedirs(path)
        except OSError as ex:
            if ex.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                logging.error("Error:", repr(ex))
                raise

    def __configure_logging(self, path_to_log_directory, log_level):
        """
        Configure logger

        :param path_to_log_directory:  path to directory to write log file in
        :return:
        """
        log_filename = datetime.datetime.now().strftime('%Y-%m-%d') + '.log'
        importer_logger = logging.getLogger('importer_logger')
        importer_logger.setLevel(log_level)
        formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(message)s')

        if not importer_logger.handlers:
            fh = logging.FileHandler(filename=os.path.join(path_to_log_directory, log_filename))
            fh.setLevel(log_level)
            fh.setFormatter(formatter)
            importer_logger.addHandler(fh)